/*******************************************************************************
clipextract-images.cpp is a sample code that demonstrates the use of 
QisMClipExtract to extract tiny clips of raster image. This file is present
in the ../samplecode directory of the QisMClipExtract product installation
------------------------------------------------------------------------------*/
#include "clipextract-images.cpp"
#include <string>
#include <vector>
#include <stdarg.h>
using namespace std;

namespace {

struct OneWindow
{
  string name;
  double lx, ly, ux, uy;
  OneWindow(): lx(0.0), ly(0.0), ux(0.0), uy(0.0) { }
};
typedef vector<OneWindow> WindowSet;

/*******************************************************************************
Client application class to work with samplecode/clipextract-images.cpp
------------------------------------------------------------------------------*/
class CQisMSampleClient: public QisMSampleClient
{
private:
  char emsg[1024];
  WindowSet windows;
  string execPath, inputFile, viewCell, layers;
  double pxX, pxY;
  int thrnumE;
  QisMClipExtractFlags::ImageFormat format;
  bool invert;

  bool Add_window(const char* windowStr_);

public: /* QisMSampleClient interface methods */
  const char* Get_executable_path() const { return this->execPath.c_str(); }
  const char* Get_input_layer_map() const;
  bool Ignore_texts() const { return true; }
  const char* Get_cad_file_path() const { return this->inputFile.c_str(); }
  const char* Get_view_cell() const { return this->viewCell.c_str(); }
  const char* Get_view_layers() const { return this->layers.c_str(); }
  int On_error(const char* fmt_, ...) { 
    if(fmt_ && fmt_[0]) {
      memset(this->emsg, 0, sizeof(this->emsg));
      va_list v;
      va_start(v, fmt_);
      vsnprintf(this->emsg, sizeof(this->emsg)-1, fmt_, v);
      va_end(v);
    }
    return 1;
  }
  const char* Get_error_msg() const { return this->emsg; }
  void On_info(const char* fmt_, ...) { 
    if(fmt_ && fmt_[0]) {
      va_list v;
      va_start(v, fmt_);
      vfprintf(stdout, fmt_, v);
      va_end(v);
    }
  }
  int Stop_file_open() const { return 0; }
  int Get_extract_window_count() const { return this->windows.size(); }
  const char* Get_extract_window(
    const int index_, double& minX_, double& minY_, double& maxX_, double& maxY_
    ) 
  {
    if((index_ >= 0) && (index_ < this->windows.size())) {
      minX_ = this->windows[index_].lx;
      minY_ = this->windows[index_].ly;
      maxX_ = this->windows[index_].ux;
      maxY_ = this->windows[index_].uy;
      return this->windows[index_].name.c_str();
    }
    return 0;
  }
  int Get_num_extract_threads() const { return this->thrnumE; }
  double Pixel_size_X() const { return this->pxX; }
  double Pixel_size_Y() const { return this->pxY; }
  bool Invert_polarity() const { return this->invert; }
  QisMClipExtractFlags::ImageFormat Image_format() const { return this->format; }
  int Process_image(
    unsigned char* bufferPtr_, const QisMClipExtractImageInfo* imageInfo_
    )
  {
    /*TODO: Process the monochome (1-bit) bitmap of 
      size = imageInfo_->Image_size() bytes
      width = (imageInfo_->Image_extents_urx() - imageInfo_->Image_extents_llx() + 1) pixels
      height = (imageInfo_->Image_extents_ury() - imageInfo_->Image_extents_lly() + 1) pixels

      bufferPtr_[0] represents the 8 pixels of the upper-left (top row) of the image
      bufferPtr_[size-1] represents the 8 pixels of the lower-right (bottom row) of the image

      This code will be executed by multiple threads at the same time. It needs
      to be have the necessary protections for multi-threaded access
    */
    fprintf(stdout, "** Here, the client application will process the raster image... **\n\n");
    return 0;
  }

public: /* Misc. methods */
  CQisMSampleClient(): thrnumE(0), format(QisMClipExtractFlags::TIFF), pxX(1.0),
  pxY(1.0), invert(false) { }
  ~CQisMSampleClient() { }
  void Print_header();
  void Print_usage(const char* argV0_);
  bool Setup(
    int argC_, char* argV_[], char* errorMsg_, const int errorMsgSz_
    );
};

}

#include <errno.h>

/* Command-line Parameters */
#define ARG_INPUTFILE "+input:"
#define OPT_CELL "-cell:"
#define OPT_LAYERS "-layers:"
#define OPT_PXSIZE "-pixelsize:" 
#define OPT_THRNUM "-thrnum:"
#define OPT_FMT_TIFF "-format:TIFF"
#define OPT_FMT_BMP "-format:BMP"
#define OPT_FMT_BUF "-format:BUF"
#define OPT_INVERT "-invert"
#define OPT_WINDOW "-window:" 
#define OPT_WINFILE "@windows:" 

#ifdef _WIN32
#define snprintf _snprintf
#define strcasecmp stricmp
#endif

const char* CQisMSampleClient::Get_input_layer_map() const
{
  /* Convert comma-separated layer list to QisLib/QisMLib input layermap 
     format */
  static string _buf;
  if((this->layers.size() == 0) || (strcasecmp(this->layers.c_str(),"All") == 0)) {
    return "ALL";
  }

  vector<char> t(this->layers.begin(), this->layers.end()); t.push_back(0);
  char* temp = &t[0];

  _buf = "ALL-NULL,";

  char* p;
  while((p = strrchr(temp,',')) != NULL)
  {
    *p = 0; p++;
    _buf += p;
    _buf += "-";
    _buf += p;
    _buf += ",";
  }

  _buf += temp;
  _buf += "-";
  _buf += temp;
  return _buf.c_str();
}

bool CQisMSampleClient::Add_window(const char* windowStr_)
{
  /* Parse window string - <basepath>,<minx>,<miny>,<maxx>,<maxy> */
  assert(windowStr_);
  memset(this->emsg, 0, sizeof(this->emsg));
  vector<char> buffer(windowStr_, windowStr_ + strlen(windowStr_) + 1);

  char* p = strtok(&buffer[0], " ,\t");
  if(!p) return true;
  string name = p;
  
  p = strtok(0, " ,\t");
  if(!p) {
    snprintf(
      this->emsg, sizeof(this->emsg)-1, "Invalid window '%s', missing minx", 
      windowStr_
      );
    return false;
  }
  double minx = atof(p);

  p = strtok(0, " ,\t");
  if(!p) {
    snprintf(
      this->emsg, sizeof(this->emsg)-1, "Invalid window '%s', missing miny", 
      windowStr_
      );
    return false;
  }
  double miny = atof(p);

  p = strtok(0, " ,\t");
  if(!p) {
    snprintf(
      this->emsg, sizeof(this->emsg)-1, "Invalid window '%s', missing maxx", 
      windowStr_
      );
    return false;
  }
  double maxx = atof(p);

  p = strtok(0, " ,\t");
  if(!p) {
    snprintf(
      this->emsg, sizeof(this->emsg)-1, "Invalid window '%s', missing maxy", 
      windowStr_
      );
    return false;
  }
  double maxy = atof(p);

  this->windows.resize(this->windows.size() + 1);
  this->windows.back().lx = minx;
  this->windows.back().ly = miny;
  this->windows.back().ux = maxx;
  this->windows.back().uy = maxy;
  this->windows.back().name = name;
  return true;
}

void CQisMSampleClient::Print_header()
{
  fprintf(stdout, "SAMPLE QISMCLIPEXTRACT CLIENT v1.0 [" __DATE__ " " __TIME__ "]\n");
  fprintf(stdout, "(c) Artwork Conversion Software Inc. (www.artwork.com)\n\n");
}

void CQisMSampleClient::Print_usage(const char* argV0_)
{
  fprintf(stdout, "%s\n", (argV0_ && argV0_) ? argV0_ : "clipextract_cpp");
  fprintf(stdout, ARG_INPUTFILE "{input-gds-oas-file}\n");
  fprintf(stdout, "[" OPT_CELL "{view-cell-name}]\n");
  fprintf(stdout, "[" OPT_LAYERS "{comma-separated-layer-list}]\n");
  fprintf(stdout, "[" OPT_PXSIZE "{pixelsize-x}[,{pixelsize-y}]]\n");
  fprintf(stdout, "[" OPT_THRNUM "{num-extract-threads}]\n");
  fprintf(stdout, "[" OPT_FMT_TIFF " | " OPT_FMT_BMP " | " OPT_FMT_BUF "]\n");
  fprintf(stdout, "[" OPT_INVERT "]\n");
  fprintf(stdout, "[" OPT_WINDOW "{base-path},{minx},{miny},{maxx},{maxy}]\n");
  fprintf(stdout, "[" OPT_WINFILE "{txt-file-one-<basepath,minx,miny,maxx,maxy>-per-line}]\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "pixelsize, window co-ordinates are in file units (e.g micron)\n");
  fprintf(stdout, "\n");
}

/* Parse command-line options */
bool CQisMSampleClient::Setup(
  int argC_, char* argV_[], char* errorMsg_, const int errorMsgSz_
  )
{
  this->Print_header();
  if(errorMsg_ && (errorMsgSz_ > 0)) memset(errorMsg_, 0, errorMsgSz_);

  if(argC_ < 3) {
    this->Print_usage((argC_ == 1) ? argV_[0] : 0);
    if(errorMsg_ && (errorMsgSz_ > 0)) {
      snprintf( 
        errorMsg_, errorMsgSz_-1, 
        "Insufficient arguments, input file and atleast one window"
        );
    }
    return false;
  }
  
  bool fullPath = true;
  #ifdef _WIN32
  fullPath = fullPath && (strchr(argV_[0], '\\') != 0);
  #else
  fullPath = fullPath && (strchr(argV_[0], '/') != 0));
  #endif
  fullPath = fullPath && (strstr(argV_[0], "..") == 0);
  if(!fullPath) {
    if(errorMsg_ && (errorMsgSz_ > 0)) {
      snprintf(
        errorMsg_, errorMsgSz_-1, "Please use fully-qualified executable path"
        );
    }
    return false;
  }
  this->execPath = argV_[0];

  vector<string> winStrs;

  for(int i=1; i<argC_; i++) {
    if(strncmp(argV_[i], ARG_INPUTFILE, sizeof(ARG_INPUTFILE)-1) == 0) {
      this->inputFile = &argV_[i][sizeof(ARG_INPUTFILE)-1];
    } else if(strncmp(argV_[i], OPT_CELL, sizeof(OPT_CELL)-1) == 0) {
      this->viewCell = &argV_[i][sizeof(OPT_CELL)-1];
    } else if(strncmp(argV_[i], OPT_LAYERS, sizeof(OPT_LAYERS)-1) == 0) {
      this->layers = &argV_[i][sizeof(OPT_LAYERS)-1];
    } else if(strncmp(argV_[i], OPT_THRNUM, sizeof(OPT_THRNUM)-1) == 0) {
      this->thrnumE = atoi(&argV_[i][sizeof(OPT_THRNUM)-1]);
      if(this->thrnumE < 0) this->thrnumE = 0;
    } else if(strcmp(argV_[i], OPT_INVERT) == 0) {
      this->invert = true;
    } else if(strcmp(argV_[i], OPT_FMT_TIFF) == 0) {
      this->format = QisMClipExtractFlags::TIFF;
    } else if(strcmp(argV_[i], OPT_FMT_BMP) == 0) {
      this->format = QisMClipExtractFlags::BMP;
    } else if(strcmp(argV_[i], OPT_FMT_BUF) == 0) {
      this->format = QisMClipExtractFlags::RBUF;
    } else if(strncmp(argV_[i], OPT_PXSIZE, sizeof(OPT_PXSIZE)-1) == 0) {
      char* p = &argV_[i][sizeof(OPT_PXSIZE)-1];
      char* s = strrchr(p, ',');
      if(s) {
        *s = 0;
        this->pxY = atof(s+1);
        this->pxX = atof(p);
        *s = ',';
      } else {
        this->pxX = this->pxY = atof(p);
      }
      if(this->pxX < 1e-6) this->pxX = 1.0;
      if(this->pxY < 1e-6) this->pxY = 1.0;
    } else if(strncmp(argV_[i], OPT_WINDOW, sizeof(OPT_WINDOW)-1) == 0) {
      winStrs.resize(winStrs.size() + 1);
      winStrs.back() = &argV_[i][sizeof(OPT_WINDOW)-1];
    } else if(strncmp(argV_[i], OPT_WINFILE, sizeof(OPT_WINFILE)-1) == 0) {
      FILE* fp = ::fopen(&argV_[i][sizeof(OPT_WINFILE)-1], "r");
      if(!fp) {
        if(errorMsg_ && (errorMsgSz_ > 0)) {
          snprintf(
            errorMsg_, errorMsgSz_-1, "Failed to open '%s', %s", 
            &argV_[i][sizeof(OPT_WINFILE)-1], strerror(errno)
            );
        }
        return false;
      }
      char line[1024];
      while(fgets(line, sizeof(line), fp)) {
        if(line[0] == '#') continue;
        char* p = strrchr(line, '\r'); if(p) *p = 0;
        p = strrchr(line, '\n'); if(p) *p = 0;
        winStrs.resize(winStrs.size() + 1);
        winStrs.back() = line;
      }
      fclose(fp);
    } else {
      if(errorMsg_ && (errorMsgSz_ > 0)) {
        snprintf(
          errorMsg_, errorMsgSz_-1, "Unrecognized argument '%s'", argV_[i]
          );
      }
      return false;
    }
  }
  
  if(this->inputFile.size() == 0) {
    if(errorMsg_ && (errorMsgSz_ > 0)) {
      snprintf(
        errorMsg_, errorMsgSz_-1, 
        "Insufficient arguments, missing input file path"
        );
    }
    return false;
  }

  if(winStrs.size() == 0) {
    if(errorMsg_ && (errorMsgSz_ > 0)) snprintf(
        errorMsg_, errorMsgSz_-1, 
        "Insufficient arguments, no windows specified"
        );
    return false;
  }
  for(int w=0, nw=winStrs.size(); w<nw; w++) {
    if(!this->Add_window(winStrs[w].c_str())) {
      if(errorMsg_ && (errorMsgSz_ > 0)) snprintf(
        errorMsg_, errorMsgSz_-1, "%s", this->emsg
        );
      return false;
    }
  }
  return true;
}

/* As required by clipextract-images.cpp */
QisMSampleClient_p Client_initialize_application(
  int argC_, char* argV_[], char* errorMsg_, const int errorMsgSz_
  )
{
  CQisMSampleClient* client = new CQisMSampleClient;
  assert(client);
  if(!client->Setup(argC_, argV_, errorMsg_, errorMsgSz_)) {
    delete client;
    return 0;
  }
  return dynamic_cast<QisMSampleClient*>(client);
}

void Client_close_application(QisMSampleClient_p handle_)
{  
  CQisMSampleClient* client = dynamic_cast<CQisMSampleClient*>(
    (QisMSampleClient*)handle_
    );
  if(client) delete client;
  fprintf(stdout, "\nEND of OPERATION\n\n");
}

